#include "para.h"
#include <sys/utsname.h> /* uname() */

/*
 * Init sockaddr. If he is null (server mode), use own IP, else use
 * that given in he.
 */
void init_sockaddr(struct sockaddr_in *addr, int port, struct hostent *he)
{
	/* host byte order */
	addr->sin_family = AF_INET;
	/* short, network byte order */
	addr->sin_port = htons(port);
	if (he)
		addr->sin_addr = *((struct in_addr *)he->h_addr);
	else
		addr->sin_addr.s_addr = INADDR_ANY;
	/* zero the rest of the struct */
	memset(&addr->sin_zero, '\0', 8);

}

/*
 * Taken from Beej's Guide to Network Programming
 *
 * Due to circumstances beyond your control, the kernel might not send
 * all the data out in one chunk, and now, my friend, it's up to us to
 * get the data out there.
 */
static int sendall(int fd, char *buf, size_t *len)
{
	int total = 0;        /* how many bytes we've sent */
	int bytesleft = *len; /* how many we have left to send */
	int n = -1;

	while (total < *len) {
		n = send(fd, buf + total, bytesleft, 0);
		if (n == -1)
			break;
		total += n;
		bytesleft -= n;
	}
	*len = total; /* return number actually sent here */

	return n == -1? -1 : 1; /* return -1 on failure, 1 on success */
}

/*
 * Send buffer of length n to socket. Log errors with givel loglevel
 */
int send_bin_buffer(int socket_fd, char *buf, size_t n)
{
	int ret;

	ret = sendall(socket_fd, buf, &n);
	if (ret < 0)
		PARA_INFO_LOG("failed to send buffer (%s)\n",
			strerror(errno));
	return ret;
}

/*
 * Send null terminated buffer to socket. Log errors with givel loglevel
 */
int send_buffer(int socket_fd, char *buf)
{
	return send_bin_buffer(socket_fd, buf, strlen(buf));
}


/* adapted from printf man page */
int send_va_buffer(int fd, char *fmt, ...)
{
	int ret;
	/* Guess we need no more than 100 bytes. */
	int n, size = 100;
	char *p = para_malloc(size);
	va_list ap;
	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf(p, size, fmt, ap);
		va_end(ap);
		/* If that worked, send the string. */
		if (n > -1 && n < size) {
			ret = send_buffer(fd, p);
			free(p);
			return ret;
		}
		/* Else try again with more space. */
		if (n > -1) /* glibc 2.1 */
			size = n + 1; /* precisely what is needed */
		else /* glibc 2.0 */
			size *= 2; /* twice the old size */
		p = para_realloc(p, size);
	}
}

int recv_bin_buffer(int socket_fd, char *buf, ssize_t len)
{

	int numbytes;

	if ((numbytes = recv(socket_fd, buf, len - 1, 0)) == -1)
		PARA_ERROR_LOG("recv error (%s)\n", strerror(errno));
	return numbytes;
}

int recv_buffer(int socket_fd, char *buf, ssize_t len)
{
	int n;

	if ((n = recv_bin_buffer(socket_fd, buf, len)) >= 0)
		buf[n] = '\0';
	return n;
}

struct hostent *get_host_info(char *host)
{
	struct hostent *he;

	if (!(he = gethostbyname(host)))
	PARA_ERROR_LOG("gethostbyname error (%s)\n", strerror(errno));
	return he;

}

int get_socket(void)
{
	int socket_fd;

	if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		PARA_ERROR_LOG("socket error (%s)\n", strerror(errno));
	return socket_fd;
}

int do_connect(int socket_fd, struct sockaddr_in *their_addr)
{
	int ret;

	if ((ret = connect(socket_fd, (struct sockaddr *)their_addr,
			sizeof(struct sockaddr))) == -1)
		PARA_ERROR_LOG("connect error (%s)\n", strerror(errno));
	return ret;
}

int do_accept(int socket_fd, struct sockaddr_in *their_addr)
{
	int new_fd;
	size_t sin_size = sizeof(struct sockaddr_in);

	if ((new_fd = accept(socket_fd, (struct sockaddr *) their_addr,
			&sin_size)) == -1)
		PARA_WARNING_LOG("%s", "accept failed\n");
	return new_fd;
}
int setserversockopts(int socket_fd)
{
	int yes = 1;

	if (setsockopt(socket_fd, SOL_SOCKET, SO_REUSEADDR, &yes,
			sizeof(int)) == -1) {
		PARA_ERROR_LOG("%s", "setsockopt failed\n");
		return -1;
	}
	return 0;
}
int do_bind(int socket_fd, struct sockaddr_in *my_addr)
{
	if (bind(socket_fd, (struct sockaddr *)my_addr,
		sizeof(struct sockaddr)) == -1) {
		PARA_EMERG_LOG("bind error (%s)\n", strerror(errno));
		return -1;
	}
	return 0;
}

int init_unix_addr(struct sockaddr_un *u, const char *name)
{
	if (strlen(name) >= UNIX_PATH_MAX)	
		return -1;
	memset(u->sun_path, 0, UNIX_PATH_MAX);
	u->sun_family = PF_UNIX;
	strcpy(u->sun_path, name);
	return 1;
}

int create_pf_socket(const char *name, struct sockaddr_un *unix_addr, int mode)
{
	int fd;

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
		return -2;
//	unlink(name);
	if (init_unix_addr(unix_addr, name) < 0)
		return -3;
//	len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family)
//		+ strlen(name) + 1;
	if (bind(fd, (struct sockaddr *) unix_addr, UNIX_PATH_MAX) < 0)
		return -3;
	if (chmod(name, mode) < 0)
		return -4;
	return fd;
}
/*
       The  code  below  passes an array of file descriptors over a Unix socket
       using SCM_RIGHTS:
*/
ssize_t pass_fd(int fd, int sock)
{
	struct msghdr msg = {0};
	struct cmsghdr *cmsg;
	int myfds[1]; /* Contains the file descriptors to pass. */
	char buf[CMSG_SPACE(sizeof myfds)];  /* ancillary data buffer */
	int *fdptr;

	myfds[0] = fd;
	msg.msg_control = buf;
	msg.msg_controllen = sizeof buf;
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_RIGHTS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(int));
	/* Initialize the payload: */
	fdptr = (int *)CMSG_DATA(cmsg);
	memcpy(fdptr, myfds, sizeof(int));
	/* Sum of the length of all control messages in the buffer: */
	msg.msg_controllen = cmsg->cmsg_len;
	return sendmsg(sock, &msg, 0);
}

ssize_t send_cred_buffer(int sock, char *buf)
{
	char control[sizeof(struct cmsghdr) + 10];
	struct msghdr   msg;
	struct cmsghdr  *cmsg;
	static struct iovec iov;
	struct ucred c;
	ssize_t ret;

	/* Response data */
	iov.iov_base = buf;
	iov.iov_len  = strlen(buf);

	c.pid = getpid();
	c.uid = getuid();
	c.gid = getgid();
	/* compose the message */
	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);

	/* attach open tty fd */
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_CREDENTIALS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct ucred));

	*(struct ucred *)CMSG_DATA(cmsg) = c;

	msg.msg_controllen = cmsg->cmsg_len;

	ret = sendmsg(sock, &msg, 0); 
	if (ret < 0)
		fprintf(stderr, "sendmsg failed\n");
	return ret;
}

static void dispose_fds(int *fds, int num)
{
	int i;

	for (i = 0; i < num; i++)
		close(fds[i]);
}

/*
 * taken from Olaf Kirch's Black Hats Manual
 */
int recv_cred_buffer(int fd, char *buffer, size_t size, struct ucred *cred)
{
	char control[255];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	int result;
	int yes = 1;

	setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &yes, sizeof(int));
	memset(&msg, 0, sizeof(msg));
	memset(buffer, 0, size);
	iov.iov_base = buffer;
	iov.iov_len = size;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	if (recvmsg(fd, &msg, 0) < 0)
		return -1;
	result = -2;
	cmsg = CMSG_FIRSTHDR(&msg);
	while (cmsg) {
		if (cmsg->cmsg_level == SOL_SOCKET
		 && cmsg->cmsg_type  == SCM_CREDENTIALS) {
			memcpy(cred, CMSG_DATA(cmsg), sizeof(*cred));
			result = iov.iov_len;
		} else
		if (cmsg->cmsg_level == SOL_SOCKET
				&& cmsg->cmsg_type  == SCM_RIGHTS) {
			dispose_fds((int *) CMSG_DATA(cmsg),
				(cmsg->cmsg_len - CMSG_LEN(0))/sizeof(int));
		}
		cmsg = CMSG_NXTHDR(&msg, cmsg);
	}
	return result;
}
char *para_hostname(void)
{
	struct utsname u;

	uname(&u);
	return para_strdup(u.nodename);
}
